#include <SerialCommand.h>
#include <MsTimer2.h>
#include <EEPROM.h>
#include <Snooze.h>
#include <TimeLib.h>
#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ST7735.h>
#include "Font_lapse.h"
#include "analogComp.h"

// Current version
#define FIRMWARE_VERSION             "Rev 4.0"
#define FIRMWARE_RELEASE_DATE        "Apr-2019"
#define FIRMWARE_BUILD               __TIME__
#define FIRMWARE_DEVELOPER1          "MACF"

// Additional colors
#define ST7735_GREY   0x528A
#define ST7735_ORANGE 0xFC00
#define ST7735_CYAN   0x07FF

// Pinout definition
#define STAT          2
#define SHUTTER       4
#define BUZZER        5
#define LCDLED        6
#define BUTTON        A0
#define NAV_BUTTON    A1
#define VOLTAGE       A3
#define CS            10  // TFT - CS & DC can use pins 2, 6, 9, 10, 15, 20, 21, 22, 23
#define DC            12  // TFT -  
#define SCLK          13  // TFT - SCLK can also use pin 14

// TFT controller definition
Adafruit_ST7735 tft = Adafruit_ST7735(CS, DC, 0); //No reset pin used

// Serial port definition
SerialCommand SCmd;

// Load Sleep drivers
SnoozeCompare compare;
SnoozeBlock config_teensy32(compare);

// Const definitions - heartbeat
const unsigned int k_heartbeat_period = 10;

// Const definitions - battery states
const unsigned int k_battery_charging = 0;
const unsigned int k_battery_not_charging = 1; //battery charge complete or not connected for charging
const unsigned int k_no_battery = 2;

// Const definitions - all events defined
const unsigned int k_dummy_event_call = 1;
const unsigned int k_analogbutton1_pressed = 2;
const unsigned int k_analogbutton2_pressed = 3;
const unsigned int k_analogbutton3_pressed = 4;
const unsigned int k_analogbutton4_pressed = 5;
const unsigned int k_analogbutton5_pressed = 6;
const unsigned int k_joystick_up = 7;
const unsigned int k_joystick_right = 8;
const unsigned int k_joystick_down = 9;
const unsigned int k_joystick_left = 10;
const unsigned int k_joystick_middle = 11;

// Const definitions - navigation joystick related
const unsigned int k_nav_button_threshold = 150;
const unsigned int k_multiple_joystick_event_delay = 25;
const unsigned int k_consecutive_events_speed_up = 10;
const unsigned int k_speed_up_delta = 10;

// Const definitions -
const unsigned int k_interval_index_lower_limit = 0;
const unsigned int k_interval_index_upper_limit = 600;
const unsigned int k_shots_index_lower_limit = 0;
const unsigned int k_shots_index_upper_limit = 45;
const unsigned int k_bulb_index_lower_limit = 0;
const unsigned int k_bulb_index_upper_limit = 100;
const unsigned int k_current_fps_index_lower_limit = 0;
const unsigned int k_current_fps_index_upper_limit = 3;
const unsigned int k_current_beep_status_lower_limit = 0;
const unsigned int k_current_beep_status_upper_limit = 1;
const unsigned int k_current_screen_off_status_lower_limit = 0;
const unsigned int k_current_screen_off_status_upper_limit = 2;
const unsigned int k_current_deep_sleep_status_lower_limit = 0;
const unsigned int k_current_deep_sleep_status_upper_limit = 2;

// Const definitions - Serial port status
const unsigned int k_serial_port_disconnected = 0;
const unsigned int k_serial_port_just_connected = 1;
const unsigned int k_serial_port_connected = 2;

// Const definition - Screen related
const int k_font_height = 10;
const int k_font_width = 7;
const int k_title_x = 2;
const int k_title_y = 3;
const int k_settings_titles_x = 4;
const int k_settings_titles_y = 20;
const int k_settings_extra_spacing_x = 2;
const int k_settings_extra_spacing_y = 4;
const int k_settings_values_x = 70;
const int k_settings_line_spacing = k_font_height + 1;
const int k_settings_values_y = k_settings_titles_y;
const int k_battery_x = 112;
const int k_battery_y = k_title_y;

// Const definitions - Screen Re-draw options
const int k_no_redraw = -1;
const int k_redraw_settings_and_values = 0;
const int k_redraw_all_values = 1;
const int k_redraw_all_values_no_title = 2;
const int k_redraw_updated_values = 3;

// Const definitions - timelapse control related
const unsigned int k_min_idle_time = 250;
const unsigned int k_interval_start = 0;
const unsigned int k_idle = 1;
const unsigned int k_bulb = 2;
const unsigned int k_aftershot = 3;

// Const definition - Battery status related
const byte k_battery_100 = 4;
const byte k_battery_75 = 3;
const byte k_battery_50 = 2;
const byte k_battery_25 = 1;
const byte k_battery_0 = 0;
const byte k_usb = 5;

// Const definition - All settings available & spare slots
enum Settings {
  //Timelapse(1) settings
  /* 01 */  k_timelapse1_mode,
  /* 02 */  k_timelapse1_interval,
  /* 03 */  k_timelapse1_shots,
  /* 04 */  k_timelapse1_bulb,
  /* 05 */  k_timelapse1_duration,
  /* 06 */  k_timelapse1_spare6,
  /* 09 */  k_timelapse1_spare7,
  /* 07 */  k_timelapse1_spare8,
  /* 08 */  k_timelapse1_spare9,
  /* 10 */  k_timelapse1_spare10,
  /* 11 */  k_timelapse1_spare11,
  /* 12 */  k_timelapse1_spare12,
  /* 13 */  k_timelapse1_spare13,
  /* 14 */  k_timelapse1_spare14,
  /* 15 */  k_timelapse1_spare15,

  //Preferences settings
  /* 16 */  k_preferences_mode,
  /* 17 */  k_preferences_beep,
  /* 18 */  k_preferences_deep_sleep,
  /* 19 */  k_preferences_screen_off,
  /* 20 */  k_preferences_spare5,
  /* 21 */  k_preferences_spare6,
  /* 22 */  k_preferences_spare7,
  /* 23 */  k_preferences_spare8,
  /* 24 */  k_preferences_spare9,
  /* 25 */  k_preferences_spare10,
  /* 26 */  k_preferences_spare11,
  /* 27 */  k_preferences_spare12,
  /* 28 */  k_preferences_spare13,
  /* 29 */  k_preferences_spare14,
  /* 30 */  k_preferences_spare15,

  //About
  /* 31 */  k_about_mode,
  /* 32 */  k_about_battery_voltage,
  /* 33 */  k_about_system_voltge,
  /* 34 */  k_about_firmware,
  /* 35 */  k_about_releasedate,
  /* 36 */  k_about_developer1,
  /* 37 */  k_about_spare7,
  /* 38 */  k_about_spare8,
  /* 39 */  k_about_spare9,
  /* 40 */  k_about_spare10,
  /* 41 */  k_about_spare11,
  /* 42 */  k_about_spare12,
  /* 43 */  k_about_spare13,
  /* 44 */  k_about_spare14,
  /* 45 */  k_about_spare15,

  //Timelapse running
  /* 46 */  k_timelapse_running_mode,
  /* 51 */  k_timelapse_running_interval,
  /* 47 */  k_timelapse_running_shots,
  /* 48 */  k_timelapse_running_remaining_shots,
  /* 52 */  k_timelapse_running_bulb,
  /* 49 */  k_timelapse_running_elapsed_time,
  /* 50 */  k_timelapse_running_remaining_time,
  /* 53 */  k_timelapse_running_spare8,
  /* 54 */  k_timelapse_running_spare9,
  /* 55 */  k_timelapse_running_spare10,
  /* 56 */  k_timelapse_running_spare11,
  /* 57 */  k_timelapse_running_spare12,
  /* 58 */  k_timelapse_running_spare13,
  /* 59 */  k_timelapse_running_spare14,
  /* 60 */  k_timelapse_running_spare15,
};

enum Modes {
  k_mode_timelapse1,
  k_mode_preferences,
  k_mode_about,
  k_mode_timelapse_running,
};

// Mode
const int k_size_of_settings = 15;
const int k_size_of_modes = 4;

char const* mode_names[] = {"<<<<  Timelapse  >>>>", "<<<<   Options   >>>>", "<<<<    About    >>>>", "   Timelapse status "};
char const* settings[][k_size_of_settings] = {
  /* Timelapse mode 1 */        {"Mode :::::", "Interval", "Shots", "Bulb", "Duration"},
  /* Preferences mode */        {"Mode :::::", "Beep", "Deep sleep", "Screen off"},
  /* About */                   {"Mode :::::", "Batt volt.", "Syst volt.", "Firmware", "Rel. date", "Developers"},
  /* Run status */              {"Mode :::::", "Interval", "Shots", "Shots 2 go", "Bulb", "tElapsed", "tRemaining" },
};

// Strings
char const* beep_labels[] = {"Off", "On"};
char const* deep_sleep_labels[] = {"Never", "1min", "2min"};
char const* screen_off_labels[] = {"Never", "1min", "2min"};
byte const fps_labels[] = {24, 30, 60, 120};

// Screen related variables
boolean refresh_screen = true;
boolean refresh_battery_status = true;
boolean program_start = true;
boolean temp_print_xy = false;
int redraw_command = k_redraw_settings_and_values;

// State machine variables
int current_mode = k_mode_timelapse1;
int current_setting = 0;
int previous_setting = -1;
int current_settings_bottom_limit = 0;
int current_settings_upper_limit = 15; ////////////////////////////////////////////////////////////////////////////////// modificar para que sea referenciado a otra variable/inicializar correctamente

// Main variables
boolean carriage_is_moving = false;
unsigned long interval_time = 50;
unsigned long interval_index = 20;
unsigned long bulb_time = 0;
unsigned long bulb_index = 4;
unsigned long aftershot_index = 5;
unsigned long shots = 10;
unsigned long shots_index = 0;
unsigned long idle_time = 40;
unsigned long shots_counter = 0;
unsigned long total_shots = 0;
unsigned long elapsed_time = 0;
unsigned long total_time = 0;
unsigned int current_system_adc_value = 0;
unsigned int current_battery_adc_value = 0;
unsigned int previous_battery_adc_value = 0;
float current_battery_voltage = 0;
float current_system_voltage = 0;
boolean t_battery_charger_a;
boolean t_battery_charger_b;
int t_battery_state;
byte current_beep_status = 0;
byte current_fps_index = 0;
unsigned int current_deep_sleep_status = 1;
unsigned int current_screen_off_status = 1;
boolean enable_deep_sleep = false;
boolean enable_screen_off = false;
unsigned long deep_sleep_counter = 0;
unsigned long screen_off_counter = 0;
unsigned int deep_sleep_delay = 0;
unsigned int screen_off_delay = 0;
unsigned int serial_port_status = k_serial_port_disconnected;

// Other variables
boolean enable_multiple_joystick_events = false;
boolean trigger_joystick_event = false;
unsigned int consecutive_events_counter = 0;
unsigned int debug_mode = 0;
volatile int analogbutton_value;
volatile unsigned int analogjoystick_value;

void setup() {
  // Pin definitions
  pinMode(SHUTTER, OUTPUT);
  pinMode(BUZZER, OUTPUT);
  pinMode(LCDLED, OUTPUT);
  pinMode(VOLTAGE, INPUT);
  pinMode(STAT, INPUT);

  // STARTUP delay
  delay(200);
  if (current_beep_status) tone(BUZZER, 1500, 50);

  // Screen setup
  digitalWrite(LCDLED, HIGH);
  tft.initR(INITR_BLACKTAB);   // initialize a ST7735S chip, black tab
  tft.fillScreen(ST7735_BLACK);
  tft.fillRect(0, 0, 128, k_font_height + 4, 0xE390);

  tft.setTextColor(ST7735_WHITE);
  tft.setCursor(k_title_x, k_title_y);
  tft.print("Lapse Controller");

  // Setup callbacks for SerialCommand commands
  SCmd.addCommand("debug0", debug0);
  SCmd.addCommand("debug1", debug1);
  SCmd.addCommand("debug2", debug2);
  SCmd.addCommand("debug3", debug3);

  // Setup button controls
  analogComparator1.setOn(1, 7); //we instruct the lib to use voltages on the pins
  CMP1_DACCR = 1 << 7 | 1 << 6 | 200; //DACEN=7 VRSEL=6
  analogComparator1.enableInterrupt(pushbutton_pressed, RISING); //we set the interrupt and when it has to be raised

  analogComparator2.setOn(0, 7); //we instruct the lib to use voltages on the pins
  CMP2_DACCR = 1 << 7 | 1 << 6 | 6; //DACEN=7 VRSEL=6
  analogComparator2.enableInterrupt(navjoystick_pressed, RISING); //we set the interrupt and when it has to be raised

  // Setup deepsleep timer
  compare.pinMode(9, HIGH, 1.65);

  // Set the Time library to use Teensy 3.0's RTC to keep time
  setSyncProvider(getTeensy3Time);

  // Heart beat - Timer interrupt
  MsTimer2::set(k_heartbeat_period, heartbeat); // 10ms period
  MsTimer2::start();
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void loop() {
  int event_id = 0;

  //Read commands
  SCmd.readSerial();

  // Check for buttons & analog joystick status
  getEvent(&event_id);
  if (event_id != 0) {
    if (enable_screen_off) {
      digitalWrite(LCDLED, HIGH);
      enable_screen_off = false;
    }
    else {
      refresh_screen = true;
      deep_sleep_counter = 0;    // Reset go to sleep counter
      screen_off_counter = 0;    // Reset screen off counter
      update_state_machine(event_id);
    }
  }

  // Refresh screen only when neccesary
  if (refresh_screen) {
    refresh_screen = false;
    draw(redraw_command);
  }

  // Refresh battery status
  if (refresh_battery_status) {
    refresh_battery_status = false;
    draw_battery_status();
  }

  // Go to Sleep
  if (enable_deep_sleep)  {
    delay(100);
    enable_deep_sleep = false;
    digitalWrite(LCDLED, LOW);
    Snooze.deepSleep( config_teensy32 );// return module that woke processor
    digitalWrite(LCDLED, HIGH);
    setSyncProvider(getTeensy3Time);
  }

  // Turn off the screen to save power
  if (enable_screen_off) {
    digitalWrite(LCDLED, LOW);
  }

  // Serial port
  if (serial_port_status == k_serial_port_just_connected) {
    initialize_serial_port();
    serial_port_status = k_serial_port_connected;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void initialize_serial_port() {

  Serial.begin(115200);
  Serial.println("The Lapse Controller Project ----");
  Serial.print("- Firmware version: ");
  Serial.println(FIRMWARE_VERSION);
  Serial.print("- Release date: ");
  Serial.println(FIRMWARE_RELEASE_DATE);
  //Serial.println("CLEARDATA");
  //Serial.println("LABEL,Acolumn,Bcolumn,...");
  //Serial.println("RESETTIMER");

  if (timeStatus() != timeSet) {
    Serial.println("- RTC: Unable to sync with the RTC");
  } else {
    Serial.println("- RTC: RTC has set the system time");
  }
  Serial.println();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//interrupt to be raised by the analog comparator
void pushbutton_pressed() {

  delay(15);
  analogbutton_value = analogRead(BUTTON);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//interrupt to be raised by the analog comparator
void navjoystick_pressed() {

  delay(15);
  analogjoystick_value = analogRead(NAV_BUTTON);
  enable_multiple_joystick_events = true;
  trigger_joystick_event = true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void update_state_machine (int event) {
  unsigned long test;

  // Event: Dummy event ----------------------------------------------------------------------------------------------------------------------
  if (event == k_dummy_event_call) {
    previous_battery_adc_value = analogRead(VOLTAGE);
    current_settings_bottom_limit = current_mode * k_size_of_settings;
    current_settings_upper_limit = (current_mode + 1) * k_size_of_settings - 1;
    current_setting = current_settings_bottom_limit;
    previous_setting = -1;
  }

  // Event: Button·1 (left) pressed while in Timelapse mode ----------------------------------------------------------------------------------
  if (event == k_analogbutton1_pressed && current_mode == k_mode_timelapse1 ) {
    current_mode = k_mode_timelapse_running;
    current_settings_bottom_limit = current_mode * k_size_of_settings;
    current_settings_upper_limit = (current_mode + 1) * k_size_of_settings - 1;
    current_setting = current_mode * k_size_of_settings + 1;
    redraw_command = k_redraw_settings_and_values;
    if (current_beep_status) tone(BUZZER, 1500, 50);
  }


  // Event: Button·2 (stop) pressed while in Timelapse_running mode --------------------------------------------------------------------------
  if (event == k_analogbutton2_pressed && current_mode == k_mode_timelapse_running) {
    current_mode = k_mode_timelapse1;
    current_settings_bottom_limit = current_mode * k_size_of_settings;
    current_settings_upper_limit = (current_mode + 1) * k_size_of_settings - 1;
    current_setting = current_settings_bottom_limit;
    previous_setting = -1;
    redraw_command = k_redraw_settings_and_values;
    if (current_beep_status) tone(BUZZER, 1500, 50);
    digitalWrite(SHUTTER, LOW);
    return;
  }


  // Other events ---------------------------------------------------------------------------------------------
  if (event == k_analogbutton2_pressed && current_mode != k_mode_timelapse_running) {
    redraw_command = k_redraw_settings_and_values;
    current_setting = current_settings_bottom_limit;
    previous_setting = -1;
    if (current_beep_status) tone(BUZZER, 2000, 50);

  }

  // Go deep sleep ---------------------------------------------------------------------------------------------
  if (event == k_joystick_middle && current_mode != k_mode_timelapse_running) {
    enable_deep_sleep = true;
  }

  // Turn off screen when in timelapse running mode-------------------------------------------------------------
  if (event == k_joystick_middle && current_mode == k_mode_timelapse_running) {
    enable_screen_off = true;
  }

  // Other events ---------------------------------------------------------------------------------------------
  if (event == k_joystick_up && current_setting >= current_settings_bottom_limit) {
    redraw_command = k_redraw_updated_values;
    previous_setting = current_setting;
    if (current_setting == current_settings_bottom_limit) {
      current_setting = current_settings_upper_limit;
    }
    else {
      current_setting -= 1;
    }
    if (current_beep_status) tone(BUZZER, 1500, 50);
  }

  if ((event == k_joystick_down) && (current_setting <= current_settings_upper_limit) ) {
    redraw_command = k_redraw_updated_values;
    previous_setting = current_setting;
    if (current_setting == current_settings_upper_limit) {
      current_setting = current_settings_bottom_limit;
    }
    else {
      current_setting += 1;
    }
    if (current_beep_status) tone(BUZZER, 1500, 50);
  }

  if (event == k_joystick_right ) {
    if (current_beep_status) tone(BUZZER, 1500, 50);
    redraw_command = k_redraw_updated_values;
    switch (current_setting) {

      case k_timelapse1_mode:
      case k_preferences_mode:
      case k_about_mode:
        if (current_mode == k_size_of_modes - 2)      current_mode = 0;
        else if (current_mode < k_size_of_modes - 2)  current_mode++;

        current_settings_bottom_limit = current_mode * k_size_of_settings;
        current_settings_upper_limit = (current_mode + 1) * k_size_of_settings - 1;
        current_setting = current_settings_bottom_limit;
        previous_setting = -1;
        redraw_command = k_redraw_settings_and_values;
        break;

      case k_timelapse_running_interval:
      case k_timelapse1_interval:
        if (consecutive_events_counter > k_consecutive_events_speed_up) {
          if (interval_index < k_interval_index_upper_limit - k_speed_up_delta) {
            interval_index += k_speed_up_delta;
          }
          else if (interval_index < k_interval_index_upper_limit) {
            interval_index++;
          }
        }
        else {
          if (interval_index < k_interval_index_upper_limit) {
            interval_index++;
          }
        }
        break;

      case k_timelapse1_bulb:
        if (bulb_index < k_bulb_index_upper_limit) {
          test = calc_bulb_value(bulb_index + 1);
          if (test < interval_time - k_min_idle_time) {
            bulb_index++;
          }
        }
        break;

      case k_timelapse_running_shots:
      case k_timelapse1_shots:
        redraw_command = k_redraw_all_values;
        if (consecutive_events_counter > k_consecutive_events_speed_up) {
          if (shots_index < k_shots_index_upper_limit - k_speed_up_delta) {
            shots_index += k_speed_up_delta;
          }
          else {
            shots_index = k_shots_index_lower_limit;
          }
        }
        else {
          if (shots_index < k_shots_index_upper_limit ) {
            shots_index++;
          }
          else {
            shots_index = k_shots_index_lower_limit;
          }
        }
        break;

      case k_timelapse1_duration:
        if (current_fps_index == k_current_fps_index_upper_limit) {
          current_fps_index = k_current_fps_index_lower_limit;
        }
        else if (current_fps_index < k_current_fps_index_upper_limit) {
          current_fps_index++;
        }
        break;

      case k_preferences_beep:
        if (current_beep_status < k_current_beep_status_upper_limit) {
          current_beep_status++;
        }
        else {
          current_beep_status = k_current_beep_status_lower_limit;
          break;
        }

      case k_preferences_deep_sleep:
        if (current_deep_sleep_status < k_current_deep_sleep_status_upper_limit) {
          current_deep_sleep_status++;
        }
        else {
          current_deep_sleep_status = k_current_deep_sleep_status_lower_limit ;
        }
        break;

      case k_preferences_screen_off:
        if (current_screen_off_status < k_current_screen_off_status_upper_limit) {
          current_screen_off_status++;
        }
        else {
          current_screen_off_status = k_current_screen_off_status_lower_limit ;
        }
        break;
    }
  }

  if (event == k_joystick_left ) {
    redraw_command = k_redraw_updated_values;
    if (current_beep_status) tone(BUZZER, 1500, 50);
    switch (current_setting) {

      case k_timelapse1_mode:
      case k_preferences_mode:
      case k_about_mode:
        if (current_mode == 0) current_mode = k_size_of_modes - 2;
        else if (current_mode > 0) current_mode--;

        current_settings_bottom_limit = current_mode * k_size_of_settings;
        current_settings_upper_limit = (current_mode + 1) * k_size_of_settings - 1;
        current_setting = current_settings_bottom_limit;
        previous_setting = -1;
        redraw_command = k_redraw_settings_and_values;
        break;

      case k_timelapse_running_interval:
      case k_timelapse1_interval:
        if (consecutive_events_counter > k_consecutive_events_speed_up) {
          if (interval_index > k_speed_up_delta) {
            test = calc_interval_value(interval_index - k_speed_up_delta);
            if (test > bulb_time + k_min_idle_time)      interval_index -= k_speed_up_delta;
          }
        }
        else {
          if (interval_index > k_interval_index_lower_limit) {
            test = calc_interval_value(interval_index - 1);
            if (test > bulb_time + k_min_idle_time)      interval_index--;
          }
        }
        break;

      case k_timelapse1_bulb:
        if (bulb_index > k_bulb_index_lower_limit) {
          bulb_index --;
        }
        break;

      case k_timelapse_running_shots:
      case k_timelapse1_shots:
        redraw_command = k_redraw_all_values;
        if (consecutive_events_counter > k_consecutive_events_speed_up) {
          if (shots_index > k_speed_up_delta)       {
            shots_index -= k_speed_up_delta;
          }
        }
        else {
          if (shots_index > k_shots_index_lower_limit)       {
            shots_index--;
          }
          else {
            shots_index = k_shots_index_upper_limit;
          }
        }
        break;

      case k_timelapse1_duration:
        if (current_fps_index == k_current_fps_index_lower_limit) {
          current_fps_index = k_current_fps_index_upper_limit;
        }
        else if (current_fps_index > k_current_fps_index_lower_limit) {
          current_fps_index--;
        }
        break;

      case k_preferences_beep:
        if (current_beep_status > k_current_beep_status_lower_limit) {
          current_beep_status--;
        }
        else {
          current_beep_status = k_current_beep_status_upper_limit;
        }
        break;

      case k_preferences_deep_sleep:
        if (current_deep_sleep_status > k_current_deep_sleep_status_lower_limit) {
          current_deep_sleep_status--;
        }
        else {
          current_deep_sleep_status = k_current_deep_sleep_status_upper_limit;
        }
        break;

      case k_preferences_screen_off:
        if (current_screen_off_status > k_current_screen_off_status_lower_limit) {
          current_screen_off_status--;
        }
        else {
          current_screen_off_status = k_current_screen_off_status_upper_limit;
        }
        break;
    }
  }

  deep_sleep_delay = current_deep_sleep_status * 6000;
  screen_off_delay = current_screen_off_status * 6000;
  interval_time = calc_interval_value(interval_index);
  bulb_time = calc_bulb_value(bulb_index);
  idle_time = interval_time - bulb_time;


  if (shots_index > 18)  {
    if (shots_index > 26)  shots = 500 * (shots_index - 25);
    else    shots = 100 * (shots_index - 17);
  }
  else {
    shots = 10 * (shots_index);
  }
  total_shots = shots;
  total_time = total_shots  * interval_time;

  if (debug_mode >= 3 && serial_port_status == k_serial_port_connected) {
    Serial.print(" Interval           ");
    Serial.print(interval_time);
    Serial.print(" | ");
    Serial.println(interval_index);
    Serial.print(" Bulb time          ");
    Serial.print(bulb_time);
    Serial.print(" | ");
    Serial.println(bulb_index);
    Serial.print(" Idle time          ");
    Serial.println(idle_time);
    Serial.println();
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void heartbeat() {
  static unsigned long beat = 0;
  static unsigned long beat_sec = 0;
  static unsigned long battery_status_refresh_counter = 0;
  static unsigned long rtc_counter = 0;
  static unsigned long t_cycle_time = 0;
  static unsigned long t_idle_time = 0;
  static unsigned long t_bulb_time = 0;
  static unsigned long t_millis_a = 0;
  static unsigned int count_for_joystick_event = 0;
  static unsigned int current_timelapse_state = 0;
  static unsigned int count_serial_port = 0;
  static unsigned int verify_serial_port_disconnected = 0;

  interrupts();

  // Serial port
  if (count_serial_port < 20) {
    count_serial_port++;
  }
  else {
    count_serial_port = 0;

    if (!Serial) {
      if ( verify_serial_port_disconnected > 5 ) {
        serial_port_status = k_serial_port_disconnected;
        verify_serial_port_disconnected = 0;
      }
      else {
        verify_serial_port_disconnected++;
      }
    }
    else if (Serial && (serial_port_status == k_serial_port_disconnected)) {
      serial_port_status = k_serial_port_just_connected;
    }
    else if (Serial && (serial_port_status == k_serial_port_connected)) {
      verify_serial_port_disconnected = 0;
    }
  }

  // RTC ------------------------------------------------------------------------------------------------------------------
  if (rtc_counter < 1000) {
    rtc_counter++;
  }
  else {
    rtc_counter = 0;
    if (debug_mode >= 3 && serial_port_status == k_serial_port_connected) {
      print_battery_voltage();
    }
  }

  // Flag to goto sleep after x seconds -----------------------------------------------------------------------------------
  if (current_deep_sleep_status > 0) {
    if ((Serial) || (current_mode == k_mode_timelapse_running)) {
      deep_sleep_counter = 0;
    }
    else {
      if (deep_sleep_counter < deep_sleep_delay) {
        deep_sleep_counter++;
      }
      else {
        deep_sleep_counter = 0;
        enable_deep_sleep = true;
      }
    }
  }

  // Flag to turn off screen
  if (current_screen_off_status > 0) {
    //if ((Serial) || (current_mode != k_mode_timelapse_running)) {
    if (current_mode != k_mode_timelapse_running) {
      screen_off_counter = 0;
    }
    else {
      if (screen_off_counter < screen_off_delay) {
        screen_off_counter++;
      }
      else {
        screen_off_counter = 0;
        enable_screen_off = true;
      }
    }
  }

  // Flag to periodically update battery status ---------------------------------------------------------------------------
  if (battery_status_refresh_counter < 150) {
    battery_status_refresh_counter++;
  }
  else {
    battery_status_refresh_counter = 0;
    refresh_battery_status = true;
  }

  // Flag to trigger additional joystick events while keeping it in one position (left, right, up, down) ------------------
  if ((enable_multiple_joystick_events == true) && (count_for_joystick_event < k_multiple_joystick_event_delay)) {
    count_for_joystick_event++;
  }
  else {
    count_for_joystick_event = 0;
    trigger_joystick_event = true;
    consecutive_events_counter++;
  }

  // Sequence for Timelapse Mode (3 states) -------------------------------------------------------------------------------
  if (current_mode == k_mode_timelapse_running)  {

    // Update screen every second
    if (beat_sec < 1000) {
      beat_sec = beat_sec + k_heartbeat_period;
    }
    else    {
      beat_sec = 0;
      redraw_command = k_redraw_all_values_no_title;
      refresh_screen = true;
    }

    // Timelapse control
    if ((shots == 0) || (shots_counter < shots)) {
      beat = beat + k_heartbeat_period;
      elapsed_time = elapsed_time + millis() - t_millis_a;
      t_millis_a = millis();

      if (beat <= idle_time) {
        if (current_timelapse_state == k_interval_start) {
          current_timelapse_state = k_idle;
          t_cycle_time = millis();
          t_idle_time = millis();
        }
        return;
      }

      if (beat < idle_time + bulb_time) {
        if (current_timelapse_state == k_idle) {
          current_timelapse_state = k_bulb;
          t_bulb_time = millis();

          digitalWrite(SHUTTER, HIGH);
        }
        return;
      }
      else {
        if (current_timelapse_state == k_bulb) {
          current_timelapse_state = k_interval_start;
          digitalWrite(SHUTTER, LOW);

          shots_counter++;
          beat = 0;

          if (debug_mode >= 3 && serial_port_status == k_serial_port_connected) {
            Serial.print(" Shot#              ");
            Serial.println(shots_counter);
            Serial.print(" Idle time(real)    ");
            Serial.print(t_bulb_time - t_idle_time);
            Serial.println(" ms");
            Serial.print(" Bulb time(real)    ");
            Serial.print(millis() - t_bulb_time);
            Serial.println(" ms");
            Serial.print(" Cycle time(real)   ");
            Serial.print(millis() - t_cycle_time);
            Serial.println(" ms");
            Serial.println("");
          }
        }
      }
    }
  }



  // Reset variables when carriage is not moving --------------------------------------------------------------------------
  if (current_mode != k_mode_timelapse_running) {
    current_timelapse_state = k_interval_start;
    beat = 0;
    beat_sec = 0;
    shots_counter = 0;
    elapsed_time = 0;
    t_millis_a = millis();
    t_cycle_time = millis();
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void draw(int mode) {
  int extra_space_y;
  int extra_space_x;

  switch  (mode) {
    case k_redraw_settings_and_values:
      tft.fillRect(0, k_settings_titles_y - 2 , 128, 160, 0);
      tft.setTextColor(ST7735_WHITE);

      for (int i = 1; i < k_size_of_settings; i++) {
        extra_space_y = (i > 0) ? k_settings_extra_spacing_y : 0;
        tft.setCursor(k_settings_titles_x, k_settings_titles_y + extra_space_y + k_settings_line_spacing * i);
        tft.print(settings[current_mode][i]);
      }

    case k_redraw_all_values:
      for (int i = 0; i < k_size_of_settings; i++) {
        extra_space_y = (i > 0) ? k_settings_extra_spacing_y : 0;
        extra_space_x = (i > 0) ? k_settings_extra_spacing_x : -1 * k_settings_values_x;
        tft.fillRect(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * i , k_font_width * 10, k_font_height, 0);
        tft.setCursor(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * i);
        tft.setTextColor(ST7735_WHITE);
        draw_settings(k_size_of_settings * current_mode + i);
      }
      tft.setTextColor(ST7735_CYAN);
      draw_settings(current_setting);

    case k_redraw_all_values_no_title:
      for (int i = 1; i < k_size_of_settings; i++) {
        extra_space_y = (i > 0) ? k_settings_extra_spacing_y : 0;
        extra_space_x = (i > 0) ? k_settings_extra_spacing_x : -1 * k_settings_values_x;
        tft.fillRect(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * i , k_font_width * 10, k_font_height, 0);
        tft.setCursor(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * i);
        tft.setTextColor(ST7735_WHITE);
        draw_settings(k_size_of_settings * current_mode + i);
      }
      tft.setTextColor(ST7735_CYAN);
      draw_settings(current_setting);

    case k_redraw_updated_values:
      int x = current_setting % k_size_of_settings;
      extra_space_y = (x > 0) ? k_settings_extra_spacing_y : 0;
      extra_space_x = (x > 0) ? k_settings_extra_spacing_x : -1 * k_settings_values_x;

      tft.fillRect(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * x , k_font_width * 10, k_font_height, 0);
      tft.setCursor(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * x);
      tft.setTextColor(ST7735_CYAN);
      draw_settings(current_setting);

      if (previous_setting >= current_settings_bottom_limit) {
        int y = previous_setting % k_size_of_settings;
        extra_space_y = (y > 0) ? k_settings_extra_spacing_y : 0;
        extra_space_x = (y > 0) ? k_settings_extra_spacing_x : -1 * k_settings_values_x;
        tft.fillRect(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * (previous_setting % k_size_of_settings), k_font_width * 10, k_font_height, 0);
        tft.setCursor(k_settings_values_x + extra_space_x, k_settings_values_y + extra_space_y + k_settings_line_spacing * (previous_setting % k_size_of_settings));
        tft.setTextColor(ST7735_WHITE);
        draw_settings(previous_setting);
      }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void draw_battery_status() {

  // Update battery & system voltage readings in the specific ADC
  current_battery_adc_value = analogRead(VOLTAGE);
  current_system_adc_value = analogRead(39); //magic number
  current_system_voltage = 1.195 * 1024 / current_system_adc_value; //magic number

  // Identify status of charging IC MCP73831
  pinMode(STAT, OUTPUT);
  digitalWrite(STAT, LOW);
  pinMode(STAT, INPUT);
  t_battery_charger_a = digitalRead(STAT);

  pinMode(STAT, OUTPUT);
  digitalWrite(STAT, HIGH);
  pinMode(STAT, INPUT);
  t_battery_charger_b = digitalRead(STAT);

  t_battery_state = t_battery_charger_a + t_battery_charger_b;

  // Debugging
  if (debug_mode == 2 && serial_port_status == k_serial_port_connected) {
    Serial.print("System voltage ");
    Serial.print(current_system_voltage);
    Serial.print("v ");
    Serial.print(current_system_adc_value);
    Serial.print(" | Battery voltage ");
    Serial.print(current_battery_voltage);
    Serial.print("v ");
    Serial.print(" | Battery Status: ");
    Serial.print(t_battery_charger_a);
    Serial.print(" ");
    Serial.print(t_battery_charger_b);
    Serial.print(" ");
    Serial.print(t_battery_state);
    Serial.print(" | Battery Curr ");
    Serial.print(current_battery_adc_value);
  }

  // Make sure that if not charging, adc values can only be lower between readings
  if (previous_battery_adc_value < current_battery_adc_value) {
    current_battery_adc_value = previous_battery_adc_value;
  }
  else {
    previous_battery_adc_value = current_battery_adc_value;
  }

  current_battery_voltage = (float)current_battery_adc_value / 158; //magic number

  // Debugging
  if (debug_mode == 2 && serial_port_status == k_serial_port_connected) {
    Serial.print(" | Prev ");
    Serial.println(previous_battery_adc_value);
  }

  // Set font to draw battery status
  tft.setFont(&Dings);
  tft.fillRect(k_battery_x, k_battery_y, 5 * k_font_width, k_font_height, 0xE390);
  tft.setCursor(k_battery_x, k_battery_y);

  // Draw battery state
  switch (t_battery_state) {
    case k_battery_charging:
      current_battery_adc_value = 5000;
      previous_battery_adc_value = current_battery_adc_value;
      tft.setTextColor(ST7735_GREEN);
      tft.print(k_usb);
      break;

    case k_battery_not_charging:
      if (current_battery_adc_value > 625) {
        tft.setTextColor(ST7735_GREEN);
        tft.print(k_battery_100);
      }
      else if (current_battery_adc_value > 610 && current_battery_adc_value <= 625) {
        tft.setTextColor(ST7735_GREEN);
        tft.print(k_battery_75);
      }
      else if (current_battery_adc_value > 540 && current_battery_adc_value <= 610) {
        tft.setTextColor(ST7735_ORANGE);
        tft.print(k_battery_50);
      }
      else if (current_battery_adc_value > 470 && current_battery_adc_value <= 540) {
        tft.setTextColor(ST7735_RED);
        tft.print(k_battery_25);
      }
      else {
        tft.setTextColor(ST7735_RED);
        tft.print(k_battery_0);
      }
      break;

      //case k_no_battery:

  }
  tft.setFont();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void draw_settings(int setting) {
  long temp_minutes;
  long temp_seconds;
  long temp_hours;
  int temp_interval;

  switch (setting) {
    case k_timelapse1_mode: //Position
    case k_preferences_mode:
    case k_about_mode:
    case k_timelapse_running_mode:
      tft.print(mode_names[current_mode]);
      break;

    case k_timelapse1_interval: //Interval
      if (interval_time < 60000) {
        tft.print(((float)interval_time / 1000), 1);
        tft.print("\"     ");
      }
      else if (interval_time < 3600000) {
        tft.print(interval_time / 60000);
        tft.print("'");
        temp_seconds = (interval_time % 60000) / 1000;
        if (temp_seconds < 10) tft.print("0");
        tft.print(temp_seconds);
        tft.print("\"     ");
      }
      else {
        tft.print(interval_time / 3600000);
        tft.print("h      ");
      }
      break;

    case k_timelapse1_bulb: //Bulb
      if (bulb_time < 60000) {
        if (bulb_time < 1000) {
          tft.print((float)bulb_time / 1000, 1);
          tft.print("\"       ");
        }
        else {
          tft.print(bulb_time / 1000);
          tft.print("\"    ");
        }
      }
      else if (bulb_time < 3600000) {
        tft.print(bulb_time / 60000);
        tft.print("'");
        temp_seconds = (bulb_time % 60000) / 1000;
        if (temp_seconds < 10) tft.print("0");
        tft.print(temp_seconds);
        tft.print("\"     ");
      }
      else {
        tft.print(bulb_time / 3600000);
        tft.print("h      ");
      }

      break;

    case k_timelapse1_shots: //Shots
      if (shots == 0) {
        tft.print("Infinite");
      }
      else {
        tft.print(shots);
      }
      break;

    case k_timelapse1_duration:
      if (shots > 0) {
        tft.print(shots / fps_labels[current_fps_index]);
        tft.print("\" @");
        tft.print(fps_labels[current_fps_index]);
      }
      else {
        tft.print("---");
      }
      break;

    case k_about_battery_voltage:
      tft.print(current_battery_voltage, 2);
      tft.print("v ");
      //tft.print(current_battery_adc_value);
      break;

    case k_about_system_voltge:
      tft.print(current_system_voltage, 2);
      tft.print("v ");
      break;


    case k_about_firmware: //Firmware version
      tft.print(FIRMWARE_VERSION);
      break;

    case k_about_releasedate: //Firmware version
      tft.print(FIRMWARE_RELEASE_DATE);
      break;

    case k_about_developer1: //Firmware version
      tft.print(FIRMWARE_DEVELOPER1);
      break;

    case  k_timelapse_running_shots:
      tft.print(shots_counter);
      if (shots > 0) {
        tft.print("|");
        tft.print(shots);
      }
      break;

    case  k_timelapse_running_remaining_shots:
      if (shots > 0) {
        tft.print(total_shots - shots_counter);
      }
      else {
        tft.print("---");
      }
      break;

    case  k_timelapse_running_elapsed_time:
      temp_hours = elapsed_time / 3600000;
      temp_minutes = (elapsed_time % 3600000) / 60000;
      temp_seconds = (elapsed_time % 60000) / 1000;
      if (temp_hours > 0) {
        if (temp_hours < 10)     tft.print("0");
        tft.print(temp_hours);
        tft.print("h");
      }
      if (temp_minutes < 10)     tft.print("0");
      tft.print(temp_minutes);
      tft.print("'");
      if (temp_seconds < 10)     tft.print("0");
      tft.print(temp_seconds);
      tft.print("\"");

      break;

    case  k_timelapse_running_remaining_time:
      if (total_time > elapsed_time) {
        temp_hours = (total_time - elapsed_time) / 3600000;
        temp_minutes = ((total_time - elapsed_time) % 3600000) / 60000;
        temp_seconds = ((total_time - elapsed_time) % 60000) / 1000;
        if (temp_hours > 0) {
          if (temp_hours < 10)     tft.print("0");
          tft.print(temp_hours);
          tft.print("h");
        }
        if (temp_minutes < 10)     tft.print("0");
        tft.print(temp_minutes);
        tft.print("'");
        if (temp_seconds < 10)     tft.print("0");
        tft.print(temp_seconds);
        tft.print("\"");
      }
      else {
        tft.print("---");
      }
      break;

    case k_timelapse_running_interval:
      temp_interval = elapsed_time % interval_time;
      if (temp_interval < 60000) {
        tft.print(temp_interval / 1000 + 1);
        tft.print("|");
      }

      if (interval_time < 60000) {
        tft.print((float)interval_time / 1000, 1);
        tft.print("\"     ");
      }
      else if (interval_time < 3600000) {
        tft.print(interval_time / 60000);
        tft.print("'");
        temp_seconds = (interval_time % 60000) / 1000;
        if (temp_seconds < 10) tft.print("0");
        tft.print(temp_seconds);
        tft.print("\"     ");
      }
      else {
        tft.print(interval_time / 3600000);
        tft.print("h      ");
      }
      break;

    case  k_timelapse_running_bulb:
      if (bulb_time < 1000) {
        tft.print((float)bulb_time / 1000, 1);
        tft.print("\"       ");
      }
      else {
        tft.print(bulb_time / 1000);
        tft.print("\" ");
        if (digitalRead(SHUTTER)) tft.print("* ");
        else tft.print("  ");
      }
      break;

    case k_preferences_beep:
      tft.print(beep_labels[current_beep_status]);
      break;

    case k_preferences_deep_sleep:
      tft.print(deep_sleep_labels[current_deep_sleep_status]);
      break;

    case k_preferences_screen_off:
      tft.print(screen_off_labels[current_screen_off_status]);
      break;

  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void getEvent(int *event_identifier) {

  // Dummy call event at program start
  if (program_start == true) {
    *event_identifier = k_dummy_event_call;
    program_start = false;
  }

  // Check if navigation buttons are kept pressed
  if ((unsigned int)analogRead(NAV_BUTTON) < k_nav_button_threshold ) {
    enable_multiple_joystick_events = false;
    analogjoystick_value = 0;
    consecutive_events_counter = 0;
  }

  // Check for Navigation button event ------------------------
  if ((trigger_joystick_event) && (analogjoystick_value > k_nav_button_threshold) )  {

    if (analogjoystick_value > 300 && analogjoystick_value < 400 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #Up: ");
      *event_identifier = k_joystick_up;
    }
    else if (analogjoystick_value > 490 && analogjoystick_value < 590 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #Left: ");
      *event_identifier = k_joystick_left;
    }
    else if (analogjoystick_value > 650 && analogjoystick_value < 750 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #Down: ");
      *event_identifier = k_joystick_down;
    }
    else if (analogjoystick_value > 810 && analogjoystick_value < 910 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #Right: ");
      *event_identifier = k_joystick_right;
    }
    else if (analogjoystick_value > 1000 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #Middle: ");
      *event_identifier = k_joystick_middle;
    }
    else {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Unrecognized button: ");

    }
    trigger_joystick_event = false;

    if (debug_mode == 1 && serial_port_status == k_serial_port_connected) {
      Serial.println(analogjoystick_value);
      *event_identifier = 0;
    }
    //analogjoystick_value = 0;
  }

  // Check for analog button event ------------------------
  if (analogbutton_value > 150)  {

    if (analogbutton_value > 300 && analogbutton_value < 500 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #1: ");
      *event_identifier = k_analogbutton1_pressed;
    }
    else if (analogbutton_value > 1000  ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #2: ");
      *event_identifier = k_analogbutton2_pressed;
    }
    /*else if (analogbutton_value > 470 && analogbutton_value < 510 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #3: ");
      event_identifier = k_analogbutton3_pressed;
      }
      else if (analogbutton_value > 460 && analogbutton_value < 540 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #4: ");
      event_identifier = k_analogbutton4_pressed;
      }
      else if (analogbutton_value > 590 && analogbutton_value < 670 ) {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) Serial.print ("Button #5: ");
      event_identifier = k_analogbutton5_pressed;
      }*/
    else {
      if (debug_mode == 1 && serial_port_status == k_serial_port_connected) {
        Serial.print ("Unrecognized button: ");
      }
    }

    if (debug_mode == 1 && serial_port_status == k_serial_port_connected) {
      Serial.println(analogbutton_value);
      *event_identifier = 0;
    }
    analogbutton_value = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
unsigned long calc_interval_value (unsigned long index) {
  unsigned long value;

  if (index <= 290)                       value = 100 * (10 + index );
  else if (index > 290 && index <= 350)   value = 500 * (index - 230);
  else if (index > 350 && index <= 410)   value = 1000 * (index - 290);
  else                                    value = 3600000 * (index - 118);

  return value;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
unsigned long calc_bulb_value (unsigned long index) {
  unsigned long value;
  if (index < 9)   value = 100 * (index + 1);
  else value = 1000 * (index - 8);

  return value;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void debug0() {
  debug_mode = 0;
  if (serial_port_status == k_serial_port_connected) {
    Serial.print("Debugging mode: ");
    Serial.print (debug_mode);
    Serial.println(" - No debugging");
  }
}

void debug1() {
  debug_mode = 1;
  if (serial_port_status == k_serial_port_connected) {
    Serial.print("Debugging mode: ");
    Serial.print (debug_mode);
    Serial.println(" - Debugging buttons");
  }
}
void debug2() {
  debug_mode = 2;
  if (serial_port_status == k_serial_port_connected) {
    Serial.print("Debugging mode: ");
    Serial.print (debug_mode);
    Serial.println(" - Debugging battery status");
  }
}

void debug3() {
  debug_mode = 3;
  if (serial_port_status == k_serial_port_connected) {
    Serial.print("Debugging mode: ");
    Serial.print (debug_mode);
    Serial.println(" - Debugging all");
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void digitalClockDisplay() {
  // digital clock display of the time
  Serial.print(hour());
  printDigits(minute());
  printDigits(second());
  Serial.print(" ");
  Serial.print(day());
  Serial.print(" ");
  Serial.print(month());
  Serial.print(" ");
  Serial.print(year());
  Serial.println();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void print_battery_voltage() {
  // digital clock display of the time
  Serial.print("DATA, TIME, TIMER, ");
  //Serial.print(hour());
  //printDigits(minute());
  //printDigits(second());
  //Serial.print(" ");
  //Serial.print(day());
  //Serial.print(" ");
  //Serial.print(month());
  //Serial.print(" ");
  //Serial.print(year());
  //Serial.print(" ");

  int a = analogRead(39);
  int mv = 1195 * 1024 / a;
  Serial.print(mv);
  Serial.print(", ");
  Serial.print(current_battery_voltage, 2);
  Serial.print(", ");
  Serial.print(current_battery_adc_value);
  Serial.println();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void printDigits(int digits) {
  // utility function for digital clock display: prints preceding colon and leading 0
  Serial.print(": ");
  if (digits < 10)
    Serial.print('0');
  Serial.print(digits);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
time_t getTeensy3Time() {
  return Teensy3Clock.get();
}

